/* --------------------------------------------------------------- *\

    file  : fbcmd.cpp
    author: m. gumz
    mail  : gumz at cs.uni-magdeburg.de

    start : Fri Jun 25 00:03:33 2004

\* --------------------------------------------------------------- */
/* --------------------------------------------------------------- *\
  
   Copyright (c) 2004 mathias gumz

  Permission is hereby granted, free of charge, to any person 
  obtaining a copy of this software and associated documentation 
  files (the "Software"), to deal in the Software without 
  restriction, including without limitation the rights to use, 
  copy, modify, merge, publish, distribute, sublicense, and/or 
  sell copies of the Software, and to permit persons to whom 
  the Software is furnished to do so, subject to the following 
  conditions:

  The above copyright notice and this permission notice shall be 
  included in all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES 
  OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
  OTHER DEALINGS IN THE SOFTWARE.

\* --------------------------------------------------------------- */

/* --------------------------------------------------------------- *\
  includes
\* --------------------------------------------------------------- */
#include <iostream>
#include <string>
#include <vector>
#include <algorithm>
#include <cstdlib>
#include <cctype>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>

/*------------------------------------------------------------------*\
\*------------------------------------------------------------------*/

#define VERSION 0.1

/*------------------------------------------------------------------*\
\*------------------------------------------------------------------*/

typedef std::vector< std::string > ArgList;

typedef struct FbInfo {

  Display* display;
  int      screen_number;
  Window   root_win;

  FbInfo() : display(0), screen_number(0), root_win(0) { };
};

typedef struct FbCmdEntry {
  
  const char* cmdname;
  int (*command)(const FbInfo& fbinfo, const ArgList& arglist);
};

/* --------------------------------------------------------------- *\
\* --------------------------------------------------------------- */

void displayUsage() {

  std::cout <<
  
"\n"
"ABOUT\n"
"\n"
"  fbcmd provides some functions to the 'core' of fluxbox\n"
"  from the commandline.\n"
"\n"
"USAGE\n"
"\n"
"  fbcmd action [arg1] [[arg2]] ...\n"
"\n"
"ACTIONS\n"
"\n"
"  help                    - display this help\n"
"  version                 - display version\n"
"  nextworkspace [int]     - change to nextworkspace\n"
"  prevworkspace [int]     - change to prevworkspace\n"
"  workspace int           - change to workspace x\n"
"\n"
"AUTHOR\n"
"\n"
"  m. gumz (akira)\n"
"\n";

}

/*------------------------------------------------------------------*\
\*------------------------------------------------------------------*/

std::string toLower(const std::string& in) {
  std::string out = in;
  std::transform(out.begin(), out.end(), out.begin(), tolower);
  return out;
}

void sendMessageToRootWindow(const FbInfo& fbinfo, Atom atom,
    unsigned long data1, unsigned long data2 = 0, unsigned long data3 = 0,
    unsigned long data4 = 0) {

  XEvent clientevent;
  clientevent.xclient.display = fbinfo.display;
  clientevent.xclient.type = ClientMessage;
  clientevent.xclient.window = fbinfo.root_win;
  clientevent.xclient.message_type = atom;
  clientevent.xclient.format = 32;
  clientevent.xclient.data.l[0] = data1;
  clientevent.xclient.data.l[1] = data2;
  clientevent.xclient.data.l[2] = data3;
  clientevent.xclient.data.l[3] = data4;

  XSendEvent(fbinfo.display,
             fbinfo.root_win,
             true,
             SubstructureRedirectMask,
             &clientevent);

}

int numberOfWorkspaces(const FbInfo& fbinfo) {
  Atom real_type;
  Atom atom = XInternAtom(fbinfo.display, 
                          "_WIN_WORKSPACE_COUNT",
                          False);
  int format;
  unsigned long num_elements, not_used;
  unsigned char *data = 0;
  if (XGetWindowProperty(fbinfo.display,
                         fbinfo.root_win,  
                         atom, 0, 1,
                         False, 0,
                         &real_type, &format, &num_elements, &not_used,  
                         (unsigned char**)&data) == Success) {
      if (data != 0)
        return static_cast<int>(*data);
  }

  return 0;
}

int numberOfCurrentWorkspace(const FbInfo& fbinfo) {
  Atom real_type;
  Atom atom = XInternAtom(fbinfo.display, 
                          "_WIN_WORKSPACE", 
                          False);
  int format;
  unsigned long num_elements, not_used;
  unsigned char *data = 0;
  if (XGetWindowProperty(fbinfo.display,
                         fbinfo.root_win,  
                         atom, 0, 1,
                         False, 0,
                         &real_type, &format, &num_elements, &not_used,  
                         (unsigned char**)&data) == Success) {
      if (data != 0)
        return static_cast<int>(*data);
  }

  return 0;
}

void changeWorkspace(const FbInfo& fbinfo, int workspace) {
  Atom atom = XInternAtom(fbinfo.display,
                          "_BLACKBOX_CHANGE_WORKSPACE",
                           False);
  sendMessageToRootWindow(fbinfo, atom, workspace);
}

/*------------------------------------------------------------------*\
\*------------------------------------------------------------------*/

int cmdHelp(const FbInfo& fbinfo, const ArgList& arglist) {
  displayUsage();
  return 0;
}

int cmdVersion(const FbInfo& fbinfo, const ArgList& arglist) {
  std::cerr << "fbcmd version " << VERSION << "\n";
  return 0;
}


int cmdWorkspace(const FbInfo& fbinfo, const ArgList& arglist) {

  if (arglist.size() < 1 ) {
    std::cerr << "error, missing argument for [workspace]. exit.\n";
    return 1;
  }

  int workspace = atoi(arglist[0].c_str());
  int current = numberOfCurrentWorkspace(fbinfo);

  if ( workspace == current )
    return 0;

  int max_workspaces = numberOfWorkspaces(fbinfo);

  if ( workspace < 0 || workspace >= max_workspaces ) {
    std::cerr << "error, wrong range for [workspace].\n" <<
                 "possible : " << 0 << " <= id < " << max_workspaces << 
                 ". exit.\n";
    return 1;
  }
  
  changeWorkspace(fbinfo, workspace);
  
  return 0;
}

int cmdNextWorkspace(const FbInfo& fbinfo, const ArgList& arglist) {

  int delta = 1;
  
  if (!arglist.empty() && atoi(arglist[0].c_str()) > 0 ) {
    delta = atoi(arglist[0].c_str());
  }

  int current = numberOfCurrentWorkspace(fbinfo);
  int max_workspaces = numberOfWorkspaces(fbinfo);

  changeWorkspace(fbinfo, (current + delta) % max_workspaces);
 
  return 0;
}

int cmdPrevWorkspace(const FbInfo& fbinfo, const ArgList& arglist) {
  
  int delta = 1;
  
  if (!arglist.empty() && atoi(arglist[0].c_str()) > 0 ) {
    delta = atoi(arglist[0].c_str());
  }

  int current = numberOfCurrentWorkspace(fbinfo);
  int max_workspaces = numberOfWorkspaces(fbinfo);

  changeWorkspace(fbinfo, (current - delta) % max_workspaces);
 
  return 0;
}


/*------------------------------------------------------------------*\
\*------------------------------------------------------------------*/

FbCmdEntry fbcommands[] = {
  { "help", cmdHelp },
  { "version", cmdVersion },
  { "workspace", cmdWorkspace },
  { "nextworkspace", cmdNextWorkspace },
  { "prevworkspace", cmdPrevWorkspace },
  { 0, 0 }
};

/*------------------------------------------------------------------*\
\*------------------------------------------------------------------*/
int main(int argc, char* argv[]) {

  unsigned int i;
  
  if ( argc < 2 ) {
    displayUsage();
    exit(0);
  }

  std::string cmd = toLower(argv[1]);

  if ( cmd == "help" ) {
    displayUsage();
    exit(0);
  }

  ArgList arglist;
  for (i = 2; i < argc; i++ ) {
    arglist.push_back(argv[i]);
  }

  if ( !(getenv("DISPLAY")) ) {
    std::cerr << "error, no DISPLAY set, exit.\n";
    exit(1);
  }

  int ret = 1;
  FbInfo fbinfo;
  fbinfo.display = XOpenDisplay(getenv("DISPLAY"));

  if ( fbinfo.display ) {
  
    fbinfo.screen_number = DefaultScreen(fbinfo.display);
    fbinfo.root_win = RootWindow(fbinfo.display, fbinfo.screen_number);

    for(i = 0; fbcommands[i].cmdname; i++ ) {

      if ( fbcommands[i].cmdname == cmd ) {
        ret = fbcommands[i].command(fbinfo, arglist);
        break;
      }
    }
    
    XCloseDisplay(fbinfo.display);
  } else {
    std::cerr << "couldnt open display\n";
    exit(1);
  }

  return ret;
}

/* --------------------------------------------------------------- *\
\* --------------------------------------------------------------- */


